%   options.max_epochs
%   options.lr_milestone = {'iteration', 'effective_lr', 'epoch'}
%   options.update_rule_within_clusters = {'positive', 'negative', 'none'}
%   options.lr_start
%   options.lr_mult                         after the learning rate milestone is reached,
%                                           learning rate is multiplied by options.lr_mult
%   options.lr_staying_mult
%   options.cross_entropy_type = {'soft', 'hard'}
%   options.weights_init_method = {'data_mean', 'data_mean_plus_random', 'kmeans', 'zero', 'load'}
%   options.display_weights = {0,1}
%   options.dataset_name ={'mnist', 'cifar'}        used for visualization

function [errTrajectory, errTrajectoryTest, W, bias]=single_layer_auto_associative_MSE(X, t, XTest, Ttest, c, options)


scale = max(X(:));
X = X / scale;
XTest = XTest / scale;

N = size(X,1);
dim = size(X,2);
M = dim;
NTest = size(XTest,1);
meanVal = mean(X);
bias=zeros(1,c*M);

if (strcmp(options.weights_init_method,'data_mean'))
    W=repmat(meanVal,[c*M,1]);
elseif (strcmp(options.weights_init_method,'small_random'))
    W= options.rand_init_std * (2*rand(c*M,dim)-1);
elseif (strcmp(options.weights_init_method,'kmeans'))
    for k=1:c
        idx = (t==k-1);
        data = X(idx,:);
        [~,centers]=kmeans (data,M,'MaxIter', 100);
        W((k-1)*M+1:k*M,:)=centers;
    end
elseif (strcmp(options.weights_init_method,'zero'))
    W=zeros(c*M,dim); 
elseif (strcmp(options.weights_init_method,'load'))
    load ('saved_initial_weights');
end

if (options.display_weights)
    handle = figure;
    %set(0,'defaultAxesFontSize', 8);
    writerObj = VideoWriter('./single_layer_auto_associative_MSE.avi');
    writerObj.FrameRate = 30;
    open(writerObj);
end

alpha = options.lr_start;
totalIter = 0;
sum_lr = 0;
iter = 0;

if (strcmp(options.lr_milestone, 'iteration'))
    target_iter = options.lr_staying_mult *M * c / alpha;
elseif (strcmp(options.lr_milestone, 'effective_lr'))
    target_sum_lr = options.lr_staying_mult *M * c;
elseif (strcmp(options.lr_milestone, 'epoch'))    
    target_iter = N;
else
    assert (0);
end

XX = repmat(X,[1,c]);
XXTest = repmat(XTest,[1,c]);
counter = 1;
errTrajectory=inf(1,options.max_epochs+1);
errTrajectoryTest=inf(1,options.max_epochs+1);
for i=1:options.max_epochs
    w = X*W' + repmat(bias,[N,1]);
    d = w - XX;
    d = d .^ 2;
    dist = zeros(N,c);
    for j=1:c
        dist(:,j) = sum(d(:,(j-1)*M+1:j*M),2);
    end
        
    [~,y] = min (dist');
    err = sum(t'~=y-1);
    err = 100 * err / N;
    errTrajectory(counter) = err;

    w = XTest*W' + repmat(bias,[NTest,1]);
    d = w - XXTest;
    d = d .^ 2;
    dist = zeros(NTest,c);
    for jj=1:c
        dist(:,jj) = sum(d(:,(jj-1)*M+1:jj*M),2);
    end

    [~,predicted_test_label] = min (dist');
    predicted_test_label = predicted_test_label - 1;
    errTest = sum(predicted_test_label ~=Ttest') * 100 / NTest;
    errTrajectoryTest(counter) = errTest;
    
    counter = counter + 1;    
    
    for j=1:N
        totalIter = totalIter + 1;
        iter = iter + 1;
        
        if (strcmp(options.lr_milestone, 'iteration'))
            if (iter >= target_iter)
                alpha = alpha * options.lr_mult;
                target_iter = options.lr_staying_mult *M * c / alpha;
                iter = 0;
            end
        elseif (strcmp(options.lr_milestone, 'effective_lr'))
            if (target_sum_lr <= sum_lr)
                alpha = alpha * options.lr_mult;
                target_sum_lr = options.lr_staying_mult *M * c;
                sum_lr = 0;
            end
        elseif (strcmp(options.lr_milestone, 'epoch'))            
            if (iter >= target_iter)
                alpha = alpha * options.lr_mult;
                iter = 0;
            end
        end
        
        if (mod(j,1000)==1)
            w = X*W' + repmat(bias,[N,1]);
            d = w - XX;
            d = d .^ 2;
            dist = zeros(N,c);
            for jj=1:c
                dist(:,jj) = sum(d(:,(jj-1)*M+1:jj*M),2);
            end

            [~,predicted_train_label] = min (dist');
            predicted_train_label = predicted_train_label - 1;
            err = sum(predicted_train_label ~= t') * 100 / N;
            
            w = XTest*W' + repmat(bias,[NTest,1]);
            d = w - XXTest;
            d = d .^ 2;
            dist = zeros(NTest,c);
            for jj=1:c
                dist(:,jj) = sum(d(:,(jj-1)*M+1:jj*M),2);
            end

            [~,predicted_test_label] = min (dist');
            predicted_test_label = predicted_test_label - 1;
            errTest = sum(predicted_test_label ~=Ttest') * 100 / NTest;
            
           
            txt = sprintf('itr: %d\t lr: %2.6f\t TrnErr: %2.2f \t TstErr: %2.2f',totalIter-1, alpha, err, errTest);
            
            if (options.display_weights==0)
                display(txt);
            else
                % Draw patterns
                if (strcmp(options.dataset_name,'mnist'))
                    wholeImage = zeros(1+29, 1+29*c,3);
                    w = 28;
                    h = 28;
                    for k=1:c
                            if (1)
                                m = 1;
                                data = X(predicted_train_label==(k-1),:);
                                if (size(data,1)>0)
                                    XMean = mean(data,1);
                                    img = reshape(XMean,[28,28])';
                                    wholeImage(1+(m-1)*29+1:1+m*29-1,1+(k-1)*29+1:1+k*29-1,1)= 1-img;
                                    wholeImage(1+(m-1)*29+1:1+m*29-1,1+(k-1)*29+1:1+k*29-1,2)= 1-img;
                                    wholeImage(1+(m-1)*29+1:1+m*29-1,1+(k-1)*29+1:1+k*29-1,3)= 1-img;
                                else
                                    img = zeros([28,28]);
                                    wholeImage(1+(m-1)*29+1:1+m*29-1,1+(k-1)*29+1:1+k*29-1,1)= 1 - img;
                                    wholeImage(1+(m-1)*29+1:1+m*29-1,1+(k-1)*29+1:1+k*29-1,2)= 1 - img;
                                    wholeImage(1+(m-1)*29+1:1+m*29-1,1+(k-1)*29+1:1+k*29-1,3)= 1 - img;
                                end
                            end
                    end
                    scale = 2;
                    %displayImage = imresize(wholeImage, [33*c*scale,33*2*M*scale]);
                    wholeImage = wholeImage/max(abs(wholeImage(:)));
                    displayImage = imresize(wholeImage, [(1+29)*scale, (1+29*c)*scale]);
                elseif (strcmp(options.dataset_name,''))
                else
                    assert(0);
                end


                imshow(displayImage,[]);
                title(txt)
                frame = getframe(handle);
                writeVideo(writerObj,frame);
            end
        end
        
        O = X(j,:)*W(t(j)*M+1:(t(j)+1)*M,:)' + bias(t(j)*M+1:(t(j)+1)*M);
        dW = alpha * (X(j,:) - O)' * X(j,:);
        dO = X(j,:)*dW';
        W(t(j)*M+1:(t(j)+1)*M,:) = W(t(j)*M+1:(t(j)+1)*M,:) + dW;    
    end
end

if (options.display_weights)
    close (writerObj);
end

w = X*W' + repmat(bias,[N,1]);
d = w - XX;
d = d .^ 2;
dist = zeros(N,c);
for jj=1:c
    dist(:,jj) = sum(d(:,(jj-1)*M+1:jj*M),2);
end

[~,y] = min (dist');
err = sum(t'~=y-1);
err = 100 * err / N;
errTrajectory(counter) = err;

w = XTest*W' + repmat(bias,[NTest,1]);
d = w - XXTest;
d = d .^ 2;
dist = zeros(NTest,c);
for jj=1:c
    dist(:,jj) = sum(d(:,(jj-1)*M+1:jj*M),2);
end

[~,predicted_test_label] = min (dist');
predicted_test_label = predicted_test_label - 1;
errTest = sum(predicted_test_label ~=Ttest') * 100 / NTest;
errTrajectoryTest(counter) = errTest;

counter = counter + 1;
end