clear variables;
close all;
dataset = 'letter';

load (sprintf('../datasets/%sTrainX', dataset));
load (sprintf('../datasets/%sTrainY', dataset));
load (sprintf('../datasets/%sTestX', dataset));
load (sprintf('../datasets/%sTestY', dataset));

TrainY = TrainY';
TestY = TestY';

N= size(TrainX,1);
X_orig = TrainX(1:N,:);
t_orig = TrainY(1:N);
p = randperm(N);
X = X_orig(p,:);
t = t_orig(p,:);

NTest= size(TestX,1);
Xtest = TestX(1:NTest,:);
Ttest = TestY(1:NTest);
close all;

c = max(max(TestY), max(TrainY))+1;
options.max_epochs = 20;
options.lr_milestone = 'epoch'; % 'effective_lr', 'iteration'
%options.lr_start = 1;
options.lr_mult  = 0.95;
options.lr_staying_mult = 10;
options.weights_init_method = 'small_random'; %{'data_mean', 'small_random', 'kmeans', 'zero', 'load'}
options.dataset_name = dataset;
options.display_weights = 0;
options.rand_init_std = 0.01;
options.adjust_for_imbalance = 0;
options.weight_decay = 0;

best_lr = 0;
best_error = inf;
err_log_lr=zeros(5,1);
for log_lr = -3:0
    options.lr_start = 10^log_lr;
    rng(11);
    sum_err = 0;
    for i = 1:5
        X_trn = [X(1:ceil((i-1)*N/5),:);X(ceil(i*N/5):end,:)];
        X_val = X(ceil((i-1)*N/5)+1:ceil(i*N/5)-1,:);
        t_trn = [t(1:ceil((i-1)*N/5));t(ceil(i*N/5):end)];
        t_val = t(ceil((i-1)*N/5)+1:ceil(i*N/5)-1);

        [~, errTrajectoryVal, ~, ~]=single_layer_auto_associative_MSE(X_trn,t_trn,X_val,t_val, c, options);
        if (errTrajectoryVal(end) == inf)
            assert (0)
        else
            sum_err = sum_err + errTrajectoryVal(end);
        end
    end
    err_log_lr(log_lr+5) = sum_err;
    if (sum_err < best_error)
        best_error = sum_err;
        best_lr = options.lr_start;
    end
end

options.lr_start = best_lr;
options.max_epochs = 200;
rng(11);
[errTrajectoryTrn, errTrajectoryTst, W, bias]=single_layer_auto_associative_MSE(X,t,Xtest,Ttest, c, options);
xval = 1:0.1:numel(errTrajectoryTrn)/10;
figure
plot (xval, 100 - errTrajectoryTrn(10:end));
xlabel('Epoch');
ylabel('Training accuracy(%)');

figure
plot (xval, 100 - errTrajectoryTst(10:end));
display( sprintf('Accuracy = %f', (100-errTrajectoryTst(end))))
xlabel('Epoch');
ylabel('Testing accuracy(%)');

save(sprintf('%s_autoassociative_MSE_experiment.mat',options.dataset_name));
