%   options.max_epochs
%   options.lr_milestone = {'iteration', 'effective_lr', 'epoch'}
%   options.update_rule_within_clusters = {'positive', 'negative', 'none'}
%   options.lr_start
%   options.lr_mult                         after the learning rate milestone is reached,
%                                           learning rate is multiplied by options.lr_mult
%   options.lr_staying_mult
%   options.cross_entropy_type = {'soft', 'hard'}
%   options.weights_init_method = {'data_mean', 'data_mean_plus_random', 'kmeans', 'zero', 'load'}
%   options.display_weights = {0,1}
%   options.dataset_name ={'mnist', 'cifar'}        used for visualization

function [errTrajectoryTrn, errTrajectoryTst, W, bias]=single_layer_competitive_cross_entropy(X, t, XTest, Ttest, c, M, options)

scale = max(X(:));
X = X / scale;
XTest = XTest / scale;

N = size(X,1);
NTest = size(XTest,1);
dim = size(X,2);
meanVal = mean(X);
bias=zeros(1,c*M);
class_weight = ones(c,1);
if (options.adjust_for_imbalance)
    for k=1:c
        n_class = sum(t==k-1);
        class_weight(k) = N / (c * n_class);
    end
end

if (strcmp(options.weights_init_method,'data_mean'))
    W=repmat(meanVal,[c*M,1]);
elseif (strcmp(options.weights_init_method,'small_random'))
    W= options.rand_init_std * (2*rand(c*M,dim)-1);
elseif (strcmp(options.weights_init_method,'kmeans'))
    for k=1:c
        idx = (t==k-1);
        data = X(idx,:);
        [~,centers]=kmeans (data,M,'MaxIter', 100);
        W((k-1)*M+1:k*M,:)=centers;
    end
elseif (strcmp(options.weights_init_method,'zero'))
    W=zeros(c*M,dim); 
elseif (strcmp(options.weights_init_method,'load'))
    load ('saved_initial_weights');
end

save('saved_initial_weights', 'W');
if (isfield(options,'initial_weight'))
    W = options.initial_weight;
end

if (isfield(options,'initial_bias'))
    bias = options.initial_bias;
end

if (options.display_weights)
    handle = figure;
    %set(0,'defaultAxesFontSize', 8);
    writerObj = VideoWriter('./single_layer_competitive_cross_entropy.avi');
    writerObj.FrameRate = 30;
    open(writerObj);
end

alpha = options.lr_start;
totalIter = 0;
sum_lr = 0;
iter = 0;

if (strcmp(options.lr_milestone, 'iteration'))
    target_iter = options.lr_staying_mult *M * c / alpha;
elseif (strcmp(options.lr_milestone, 'effective_lr'))
    target_sum_lr = options.lr_staying_mult *M * c;
elseif (strcmp(options.lr_milestone, 'epoch'))    
    target_iter = N;
else
    assert (0);
end
counter = 1;
errTrajectoryTrn=inf(1,options.max_epochs*10+1);
errTrajectoryTst=inf(1,options.max_epochs*10+1);
for i=1:options.max_epochs    
    for j=1:N
        if (mod (j, N/10) <= 1)
            w = X*W' + repmat(bias,[N,1]);
            [~,l]=max(w');
            y = floor((l-1)/M);
            err = sum(t'~=y);
            err = 100 * err / N;
            errTrajectoryTrn(counter) = err;
            
            w = XTest*W' + repmat(bias, [NTest,1]);
            [~,l] = max(w');
            y     = floor((l-1)/M);
            errTest = sum(Ttest'~=y);
            errTest = 100 * errTest / NTest;
            errTrajectoryTst(counter) = errTest;
            
            counter = counter + 1;            
        end
        
        totalIter = totalIter + 1;
        iter = iter + 1;
        
        if (strcmp(options.lr_milestone, 'iteration'))
            if (iter >= target_iter)
                alpha = alpha * options.lr_mult;
                target_iter = options.lr_staying_mult *M * c / alpha;
                iter = 0;
            end
        elseif (strcmp(options.lr_milestone, 'effective_lr'))
            if (target_sum_lr <= sum_lr)
                alpha = alpha * options.lr_mult;
                target_sum_lr = options.lr_staying_mult *M * c;
                sum_lr = 0;
            end
        elseif (strcmp(options.lr_milestone, 'epoch'))            
            if (iter >= target_iter)
                alpha = alpha * options.lr_mult;
                iter = 0;
            end
        end
        
        if (mod(j,100000)==1)
            w = X*W' + repmat(bias,[N,1]);
            [~,l]=max(w');
            y = floor((l-1)/M);
            err = sum(t'~=y);
            err = 100 * err / N;
            predicted_train_label = l;
                        
            w = XTest*W' + repmat(bias, [NTest,1]);
            [~,l] = max(w');
            y     = floor((l-1)/M);
            errTest = sum(Ttest'~=y);
            errTest = 100 * errTest / NTest;

           
            txt = sprintf('itr: %d\t lr: %2.6f\t TrnErr: %2.2f \t TstErr: %2.2f',totalIter-1, alpha, err, errTest);
            
            if (options.display_weights==0)
                display(txt);
            else
                % Draw patterns
                skip = 0;
                if (strcmp(options.dataset_name,'mnist'))
                    wholeImage = zeros(1+29*M, 1+29*c,3);
                    w = 28;
                    h = 28;
                    for k=1:c
                        for m=1:M
                            if (1)
                                data = X(predicted_train_label==(k-1)*M+m,:);
                                if (size(data,1)>0)
                                    XMean = mean(data,1);
                                    img = reshape(XMean,[28,28])';
                                    wholeImage(1+(m-1)*29+1:1+m*29-1,1+(k-1)*29+1:1+k*29-1,1)= 1-img;
                                    wholeImage(1+(m-1)*29+1:1+m*29-1,1+(k-1)*29+1:1+k*29-1,2)= 1-img;
                                    wholeImage(1+(m-1)*29+1:1+m*29-1,1+(k-1)*29+1:1+k*29-1,3)= 1-img;
                                else
                                    img = zeros([28,28]);
                                    wholeImage(1+(m-1)*29+1:1+m*29-1,1+(k-1)*29+1:1+k*29-1,1)= 1 - img;
                                    wholeImage(1+(m-1)*29+1:1+m*29-1,1+(k-1)*29+1:1+k*29-1,2)= 1 - img;
                                    wholeImage(1+(m-1)*29+1:1+m*29-1,1+(k-1)*29+1:1+k*29-1,3)= 1 - img;
                                end
                                
%                                 img = reshape(W((k-1)*M+m,:),[28,28])';
%                                 MVal = max(abs(img(:)))+0.000001;
%                                 img = img / MVal;
%                                 wholeImage((k-1)*29+1:k*29-1,(M+m-1)*29+1:(M+m)*29-1,1)= max(img,0);
%                                 wholeImage((k-1)*29+1:k*29-1,(M+m-1)*29+1:(M+m)*29-1,2)= max(-img,0);                                
                            elseif (0)
                                img = reshape(W((k-1)*M+m,:),[28,28])';
                                MVal = max(abs(img(:)))+0.000001;
                                img = img / MVal;
                                wholeImage((k-1)*29+1:k*29-1,(m-1)*29+1:m*29-1,1)= max(img,0);
                                wholeImage((k-1)*29+1:k*29-1,(m-1)*29+1:m*29-1,2)= max(-img,0);

                                %                             img = reshape(W((k-1)*M+m,:),[28,28])';
                                %                             wholeImage((k-1)*29+1:k*29-1,29*2*M+(m-1)*29+1:29*2*M+m*29-1)=img;
                            else
                                img = reshape(W((k-1)*M+m,:),[28,28])';
                                MVal = max(abs(img(:)))+0.000001;
                                img = img / MVal;
                                img2 = imadjust(img,stretchlim(img),[]);
                                wholeImage((m-1)*29+1:m*29-1,(k-1)*29+1:k*29-1)= img2;
                            end
                        end
                    end
                    scale = 2;
                    %displayImage = imresize(wholeImage, [33*c*scale,33*2*M*scale]);
                    wholeImage = wholeImage/max(abs(wholeImage(:)));
                    displayImage = imresize(wholeImage, [(1+29*M)*scale, (1+29*c)*scale]);
                elseif (strcmp(options.dataset_name,'cifar10'))
                    wholeImage = zeros(33*M,33*c,3);
                    w = 32;
                    h = 32;
                    channel = 3;
                    hp1 = h + 1;
                    wp1 = w + 1;
                    for k=1:c
                        for m=1:M
                            %                         wholeImage((m-1)*hp1+1:m*hp1-1,(k-1)*wp1+1:k*wp1-1,:)=permute(reshape(W((k-1)*M+m,:),[w,h,channel]),[2,1,3]);
                            %                         wholeImage(hp1*M+(m-1)*hp1+1:hp1*M+m*hp1-1,(k-1)*wp1+1:k*wp1-1,:)=permute(reshape(WNeg((k-1)*M+m,:),[w,h,channel]),[2,1,3]);
                            if(1)
                                %                             wholeImage((k-1)*wp1+1:k*wp1-1,(m-1)*hp1+1:m*hp1-1,:)=reshape(W((k-1)*M+m,:),[w,h,channel]);
                                %                             wholeImage((k-1)*wp1+1:k*wp1-1,hp1*M+(m-1)*hp1+1:hp1*M+m*hp1-1,:)=reshape(WNeg((k-1)*M+m,:),[w,h,channel]);
                                wholeImage((m-1)*hp1+1:m*hp1-1,(k-1)*wp1+1:k*wp1-1,:)=reshape(W((k-1)*M+m,:),[w,h,channel]);
                                wholeImage(hp1*M+(m-1)*hp1+1:hp1*M+m*hp1-1,(k-1)*wp1+1:k*wp1-1,:)=reshape(WNeg((k-1)*M+m,:),[w,h,channel]);
                            else
                                img = reshape(W((k-1)*M+m,:),[w,h,channel]);
                                MVal = max(abs(img(:)))+0.000001;
                                img = img / MVal;
                                img2 = imadjust(img,stretchlim(img),[]);
                                wholeImage((k-1)*wp1+1:k*wp1-1,(m-1)*hp1+1:m*hp1-1,:)=img2;

                                img = reshape(WNeg((k-1)*M+m,:),[w,h,channel]);
                                MVal = max(abs(img(:)))+0.000001;
                                img = img / MVal;
                                img2 = imadjust(img,stretchlim(img),[]);
                                wholeImage((k-1)*wp1+1:k*wp1-1,hp1*M+(m-1)*hp1+1:hp1*M+m*hp1-1,:)=img2;
                            end
                            %wholeImage(hp1*2*M+(m-1)*hp1+1:hp1*2*M+m*hp1-1,(k-1)*wp1+1:k*wp1-1,:)=permute(reshape(W((k-1)*M+m,:),[w,h,channel]),[2,1,3]);
                        end
                    end
                    scale = 4;
                    displayImage = imresize(wholeImage, [hp1*M*scale,wp1*c*scale]);
                elseif (strcmp(options.dataset_name,''))
                else
                    %assert(0);
                    skip = 1;
                end

                if (skip == 0)
                    imshow(displayImage,[]);
                    title(txt)
                    frame = getframe(handle);
                    writeVideo(writerObj,frame);
                end
            end
        end
        
        w = W*X(j,:)' + bias';
        w = w - max(w);
        z = exp(w);
        denum = sum(z);
        if denum == 0
            denum = 10e-10;
        end
        z = z / denum;
        
        tau = zeros(size(z));
        
        if (strcmp(options.cross_entropy_type, 'soft'))
            denum = (sum (z(t(j)*M+1:(t(j)+1)*M)));
            if (denum ==0)
                denum = 10e-10;
            end
            tau(t(j)*M+1:(t(j)+1)*M) = z(t(j)*M+1:(t(j)+1)*M) / denum;
                
            
        elseif (strcmp(options.cross_entropy_type, 'hard'))
            [~,q]     = max(z(t(j)*M+1:(t(j)+1)*M));
            tj_q      = t(j)*M + q;
            tau(tj_q) = 1;
        else
            assert(0);
        end
                
        [~,winner] = max (z);
        y = floor((winner-1)/M);
        
        dE_dwk = class_weight(t(j)+1) * (z - tau);
        sum_lr = sum_lr + sum(abs(dE_dwk(:)));
        bias = bias - alpha * dE_dwk';
        if sum (isnan(dE_dwk(:))) > 0
            nanproblem = 1;
        end
        W = W - alpha * (dE_dwk * X(j,:) + options.weight_decay * W);    
    end
end

if (options.display_weights)
    close (writerObj);
end

w = X*W' + repmat(bias,[N,1]);
[~,l]=max(w');
y = floor((l-1)/M);
err = sum(t'~=y);
err = 100 * err / N;
errTrajectoryTrn(counter) = err;

w = XTest*W' + repmat(bias, [NTest,1]);
[~,l] = max(w');
y     = floor((l-1)/M);
errTest = sum(Ttest'~=y);
errTest = 100 * errTest / NTest;
errTrajectoryTst(counter) = errTest;

counter = counter + 1;
end